from __future__ import annotations

import os
from typing import Any

import typer
from ase.db import connect
from ase.db.core import check
from ase.gui.gui import GUI, Images
from rich.panel import Panel
from textual import on, work
from textual._two_way_dict import TwoWayDict
from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.containers import Container
from textual.coordinate import Coordinate
from textual.driver import Driver
from textual.reactive import var
from textual.widgets import Footer, Header, Input
from textual.widgets._data_table import ColumnKey, StringKey
from textual.worker import Worker, WorkerState
from typer.rich_utils import (
    ALIGN_ERRORS_PANEL,
    ERRORS_PANEL_TITLE,
    STYLE_ERRORS_PANEL_BORDER,
    _get_rich_console,
)
from typing_extensions import Annotated

from texase.addcolumn import AddColumnBox
from texase.data import ALL_COLUMNS, ASEReadError, ASEWriteError, instantiate_data
from texase.details import Details
from texase.edit import AddBox, EditBox
from texase.files_io import FilesIOScreen
from texase.filter import FilterBox
from texase.formatting import (
    convert_value_to_int_float_or_bool,
    correctly_typed_kvp,
    format_value,
    kvp_exception,
)
from texase.help import HelpScreen
from texase.keys import KeyBox
from texase.search import Search
from texase.table import TexaseTable
from texase.yesno import YesNoScreen


class TEXASE(App):
    BINDINGS = [
        Binding("ctrl+g", "hide_all", "Hide all boxes", show=False),
        Binding("ctrl+z", "suspend_process", "Suspend the app", show=False),
    ]

    CSS_PATH = "texase.tcss"

    # variables that are watched. Remember also to add them to the
    # action_hide_all function
    show_details = var(False)
    show_add_column_box = var(False)
    show_search_box = var(False)
    show_filter = var(False)
    show_edit = var(False)
    show_add_kvp = var(False)

    def __init__(self, path: str) -> None:
        self.path = path
        self.sort_columns: list[str] = ["id"]
        self.sort_reverse: bool = False
        self.gui: GUI | None = None
        super().__init__()

    def compose(self) -> ComposeResult:
        """Called to add widgets to the app."""
        yield Header()
        yield Footer()
        yield MiddleContainer(
            # Boxes put centered at the top, but in a higher layer, of the table
            FilterBox(id="filter-box", classes="topbox"),
            # Boxes docked at the bottom in the same layer of the table
            Search(
                Input(id="search-input", placeholder="Search.."),
                classes="bottombox",
                id="search-box",
            ),
            EditBox(id="edit-box", classes="bottombox"),
            AddBox(id="add-kvp-box", classes="bottombox"),
            AddColumnBox(id="add-column-box", classes="bottombox"),
            # Other boxes
            Details(id="details"),
            TexaseTable(id="table"),
            KeyBox(id="key-box"),
        )

    async def on_mount(self) -> None:
        # Table
        table = self.query_one(TexaseTable)

        self.load_initial_data(table)

        table.focus()

        self.finish_mounting()

    @work
    async def finish_mounting(self) -> None:
        # Load the rest of the data
        table = self.query_one(TexaseTable)

        key_box = self.query_one(KeyBox)
        key_box.loading = True

        self.load_remaining_data(table)

    async def on_worker_state_changed(self, event: Worker.StateChanged) -> None:
        """If load remaining data is done, stop loading, notify and populate key box."""
        if (
            event.worker.name == "load_remaining_data"
            and event.worker.state == WorkerState.SUCCESS
        ):
            key_box = self.query_one(KeyBox)
            key_box.loading = False

            # TODO: check that additional keys are added after initial load
            await self.populate_key_box()  # type: ignore

            self.notify("Loading Done!", severity="information", timeout=1.5)

    def load_initial_data(self, table: TexaseTable) -> None:
        # db data
        data = instantiate_data(self.path, limit=100)
        self.data = data

        table.populate_table(data)

    @work(thread=True)
    def load_remaining_data(self, table: TexaseTable) -> None:
        indices = self.data.add_remaining_rows_to_df()
        if len(indices) > 0:
            self.call_from_thread(table.add_table_rows, self.data, indices)

    async def populate_key_box(self) -> None:
        key_box = self.query_one(KeyBox)
        await key_box.populate_keys(self.data.unused_columns())

    def remove_filter_from_table(self, filter_tuple: tuple) -> None:
        self.query_one(TexaseTable).remove_filter(*filter_tuple)

    # Import / Export

    @work
    async def action_export_rows(self) -> None:
        """Export the marked rows or selected row of the table to a file"""
        table = self.query_one(TexaseTable)
        ids = table.ids_to_act_on()

        # Show the directory tree with an input box to select a file
        # name and location
        output_file = await self.push_screen_wait(FilesIOScreen(False))
        if output_file is not None:
            try:
                self.data.export_rows(ids, output_file)
            except ASEWriteError as e:
                self.notify_error(
                    f"ASE cannot write to the file {output_file}, it gives:\n {e}",
                    "ASE write error",
                    timeout=5,
                )

    @work
    async def action_import_rows(self) -> None:
        """Export the marked rows or selected row of the table to a file"""
        input_file = await self.push_screen_wait(FilesIOScreen(True))
        if input_file is not None:
            # Put info in db and data.df
            try:
                added_indices = self.data.import_rows(input_file)
            except ASEReadError as e:
                self.notify_error(
                    f"ASE cannot read the file {input_file}, it gives:\n {e}",
                    "ASE read error",
                    timeout=5,
                )
            else:
                # Clear and reoccupy the table
                table = self.query_one(TexaseTable)
                table.loading = True
                table.add_table_rows(self.data, indices=added_indices)

                # Update KeyBox
                await self.populate_key_box()

                table.loading = False
                table.focus()

    # Sorting
    def action_sort_column(self) -> None:
        # Get the highlighted column
        table = self.query_one(TexaseTable)
        self.sort_table(table.column_at_cursor(), table)

    def on_data_table_header_selected(
        self, selected: TexaseTable.HeaderSelected
    ) -> None:
        table = selected.data_table
        col_name = str(selected.label)
        self.sort_table(col_name, table)

    def sort_table(self, col_name: str, table: TexaseTable) -> None:
        # Save the row key of the current cursor position
        row_key = table.coordinate_to_cell_key(Coordinate(table.cursor_row, 0)).row_key

        # Sort the table
        ordered_index = self.data.sort(col_name)

        table._row_locations = TwoWayDict(
            {
                StringKey(str(key)): new_index
                for new_index, key in enumerate(ordered_index)
            }
        )
        table._update_count += 1
        table.refresh()

        # After finished sort make the cursor go to the same cell as before sorting
        table.cursor_coordinate = Coordinate(
            table._row_locations.get(row_key), table.cursor_column
        )

        # How sort does it:
        # self._row_locations = TwoWayDict(
        #     {key: new_index for new_index, (key, _) in enumerate(ordered_rows)}
        # )
        # self._update_count += 1
        # self.refresh()
        # table.sort(*self.sort_columns, reverse=self.sort_reverse)

    # Details sidebar
    def action_toggle_details(self) -> None:
        self.show_details = not self.show_details
        table = self.query_one(TexaseTable)
        if self.show_details:
            # Get the highlighted row
            row_id = table.row_id_at_cursor()
            details = self.query_one(Details)
            details.clear_modified_and_deleted_keys()
            details.update_kvplist(*self.data.row_details(row_id))
            details.update_data(self.data.row_data(row_id))

            # Set focus on the details sidebar
            details.set_focus()

        else:
            # Set focus back on the table
            table.focus()

    def save_details(
        self, key_value_pairs: dict, data: dict, deleted_keys: set
    ) -> None:
        """Called when the user calls save in the details sidebar."""

        # Add deleted keys to key_value_pairs with None values
        for key in deleted_keys:
            key_value_pairs[key] = None

        table = self.query_one(TexaseTable)
        table.update_row_editable_cells(key_value_pairs)

        row_id = table.row_id_at_cursor()
        self.data.update_row(row_id, key_value_pairs, data)

    def watch_show_details(self, show_details: bool) -> None:
        """Called when show_details is modified."""
        dv = self.query_one(Details)
        dv.display = show_details

    def action_hide_all(self) -> None:
        self.show_details = False
        self.show_add_column_box = False
        self.show_search_box = False
        self.show_filter = False
        self.show_edit = False
        self.show_add_kvp = False
        self.query_one(TexaseTable).focus()

    # Help screen
    def action_toggle_help(self) -> None:
        self.push_screen(HelpScreen())

    # Add/Delete key-value-pairs
    def action_add_key_value_pair(self) -> None:
        table = self.query_one(TexaseTable)
        self.show_add_kvp = True
        addbox = self.query_one("#add-kvp-box", AddBox)
        table.update_add_box(addbox)
        addbox.focus()

    def watch_show_add_kvp(self, show_add_kvp: bool) -> None:
        box = self.query_one("#add-kvp-box")
        box.display = show_add_kvp

    @work
    async def action_delete_key_value_pairs(self) -> None:
        table = self.query_one(TexaseTable)
        if not table.is_cell_editable(uneditable_columns=ALL_COLUMNS):
            return
        question = table.delete_kvp_question()
        if await self.push_screen_wait(YesNoScreen(question)):
            table.delete_selected_key_value_pairs()

            # Remove in db and df
            column_name = table.column_at_cursor()
            self.data.update_value(
                table.ids_to_act_on(), column=column_name, value=None
            )

        # If no other key value pairs are present in the column, delete the column from the table
        self.data.clean_user_keys()
        table.check_columns(self.data)
        await self.query_one(KeyBox).populate_keys(self.data.unused_columns())

    # Delete rows
    @work
    async def action_delete_rows(self) -> None:
        """Delete the currently marked rows."""
        table = self.query_one(TexaseTable)
        if await self.push_screen_wait(YesNoScreen(table.delete_row_question())):
            # Remove in db and df
            self.data.delete_rows_from_df_and_db(table.ids_to_act_on())

            # Then remove in table
            table.delete_selected_rows()

    # Edit
    def action_edit(self) -> None:
        table = self.query_one(TexaseTable)
        if table.is_cell_editable():
            self.show_edit = True
            editbox = self.query_one("#edit-box", EditBox)
            table.update_edit_box(editbox)
            editbox.focus()

    def watch_show_edit(self, show_edit: bool) -> None:
        editbox = self.query_one("#edit-box")
        editbox.display = show_edit

    # Search
    def action_search(self) -> None:
        # show_search_box is set to True since the search bar is able
        # to close itself after a search.
        self.show_search_box = True
        search_input = self.query_one("#search-input")
        search_input.focus()  # This is the input box

        search = self.query_one(Search)
        search._table = self.query_one(TexaseTable)
        search._data = self.data
        search_input.value = ""
        search.set_current_cursor_coordinate()

    def watch_show_search_box(self, show_search_box: bool) -> None:
        searchbar = self.query_one(Search)
        searchbar.display = show_search_box

    # Filter
    async def action_filter(self) -> None:
        self.show_filter = True
        filterbox = self.query_one("#filter-box", FilterBox)
        await filterbox.focus_filterbox()

    def watch_show_filter(self, show_filter: bool) -> None:
        searchbar = self.query_one("#filter-box")
        searchbar.display = show_filter

    # Column action
    def action_add_column(self) -> None:
        self.show_add_column_box = True
        self.query_one("#add-column-box").focus()

    def watch_show_add_column_box(self, show_box: bool) -> None:
        addcolumnbox = self.query_one(AddColumnBox)
        addcolumnbox.display = show_box

    async def action_remove_column(self) -> None:
        """Remove the column that the cursor is on.

        Also remove the column from chosen_columns."""

        table = self.query_one(TexaseTable)
        # Save the name of the column to remove
        cursor_column_index = table.cursor_column
        column_to_remove = str(table.ordered_columns[cursor_column_index].label)

        # Add the column to the KeyBox
        await self.query_one(KeyBox).add_key(column_to_remove)

        self.remove_column_from_table(column_to_remove)

    def remove_column_from_table(self, column_name: str) -> None:
        table = self.query_one(TexaseTable)
        # Remove the column from the table in data
        self.data.remove_from_chosen_columns(column_name)

        # col_key = table.ordered_columns[cursor_column_index].key
        col_key = ColumnKey(column_name)
        table.remove_column(col_key)

    def action_view(self) -> None:
        """View the currently selected images, if no images are
        selected then view the row the cursor is on"""
        table = self.query_one(TexaseTable)
        if table.marked_rows:
            images = [self.data.get_atoms(id) for id in table.get_marked_row_ids()]
        else:
            images = [self.data.get_atoms(table.row_id_at_cursor())]
        self.gui = GUI(Images(images))
        # Only run if we are not doing a pytest
        if "PYTEST_CURRENT_TEST" not in os.environ:
            self.gui.run()

    def add_column_to_table_and_remove_from_keybox(self, column: str) -> None:
        """Add a column to the table and remove it from the KeyBox."""
        table = self.query_one(TexaseTable)
        self.data.add_to_chosen_columns(column)
        table.add_column_and_values(column)
        self.query_one(KeyBox).remove_key(column)

    def on_input_submitted(self, submitted: Input.Submitted):
        table = self.query_one(TexaseTable)
        if (
            submitted.validation_result is not None
            and not submitted.validation_result.is_valid
        ):
            return
        if submitted.control.id == "add-column-input":
            self.add_column_to_table_and_remove_from_keybox(submitted.value)

            self.show_add_column_box = False
            table.focus()
        elif submitted.control.id == "edit-input":
            column = table.column_at_cursor()
            if column == "pbc":
                value = submitted.value.upper()
            else:
                value = convert_value_to_int_float_or_bool(submitted.value)

            exception_from_kvp = kvp_exception(column, value)
            if exception_from_kvp is not None:
                self.notify_error(exception_from_kvp, "ValueError")
                return

            # kvp is ok, but give a warning if type changed
            self.notify_if_kvp_type_changed(
                column, value, self.data.df.loc[table.cursor_row, column]
            )

            # Update table
            table.update_cell_from_edit_box(format_value(value))

            # Update data
            self.data.update_value(
                ids=table.row_id_at_cursor(),
                column=table.column_at_cursor(),
                value=value,
            )

            # Go back to original view
            self.show_edit = False
            table.focus()

        elif submitted.control.id == "add-input":
            key, value = correctly_typed_kvp(submitted.value)

            exception_from_kvp = kvp_exception(key, value)
            if exception_from_kvp is not None:
                self.notify_error(exception_from_kvp, "ValueError")
                return

            # Add to chosen_columns and user_keys in data
            self.data.add_new_user_key(key, show=True)

            # Update data
            if table.marked_rows:
                # If there are marked rows, add the key value pair to
                # all marked rows
                self.data.update_value(
                    ids=table.get_marked_row_ids(), column=key, value=value
                )
            else:
                # If there are no marked rows, add the key value pair
                # to the row the cursor is on
                self.data.update_value(
                    ids=table.row_id_at_cursor(), column=key, value=value
                )

            # Update table
            table.update_cell_from_add_box(key, format_value(value))

            # Go back to original view
            self.show_add_kvp = False
            table.focus()

    def notify_if_kvp_type_changed(
        self, key: str, new_value: Any, old_value: Any
    ) -> None:
        if not isinstance(new_value, type(old_value)):
            self.notify(
                f"The type of [bold]{key} = {new_value}[/bold] has changed from "
                + f"[bold]{type(old_value).__name__}[/bold] to [bold]{type(new_value).__name__}[/bold].",
                severity="warning",
            )

    # Shouldn't this be deleted??
    def is_kvp_valid(self, key, value):
        """Check that key-value-pair is valid for ase.db

        It is ok to edit pbc, we make this check first."""

        if key == "pbc":
            try:
                check_pbc_string_validity(value)
            except ValueError as e:
                self.notify_error(str(e), "ValueError")
                return False
            return True

        try:
            check({key: value})
        except ValueError as e:
            # Notify that the key-value-pair is not valid with the
            # raised ValueError and then return
            self.notify_error(str(e), "ValueError")
            return False
        return True

    def notify_error(self, e: str, error_title: str, timeout: float = 3) -> None:
        self.notify(
            e,
            severity="error",
            title=error_title,
            timeout=timeout,
        )

    def action_quit(self) -> None:
        self.data.save_chosen_columns()
        super().exit()

    def action_suspend_process(self) -> None:
        self.data.save_chosen_columns()
        super().action_suspend_process()

    @on(Driver.SignalResume)
    @work
    async def action_update_view(self) -> None:
        """Check if the db has been updated since it was last read.

        If so update the table."""
        # if not self.data.is_df_up_to_date():
        remove_idx, update_idx, add_idx = self.data.updates_from_db()

        table = self.query_one(TexaseTable)
        table.delete_rows([table.row_index_to_row_key(idx) for idx in remove_idx])

        table.add_table_rows(self.data, add_idx)
        table.update_table_rows(self.data, update_idx)

        table.check_columns(self.data)

        # Update the KeyBox
        await self.populate_key_box()


def check_pbc_string_validity(string):
    """Check if the string is a valid pbc string. I.e. on the form TTT, TFT, FFF, etc."""
    # check if the string has exactly three characters
    if len(string) == 3:
        # convert the string to upper case
        string = string.upper()
        # loop through each character in the string
        for char in string:
            # check if the character is either t or f
            if char not in ["T", "F"]:
                # raise a ValueError with a descriptive message
                raise ValueError(f"{string} contains characters that are not T or F!")
        # return True if all characters are t or f
        return True
    else:
        # raise a ValueError with a descriptive message
        raise ValueError(f"{string} does not have exactly three characters!")


def is_db_empty(db_path: str) -> bool:
    """Check if the database is empty."""
    # Quick test of a db file. If it doesn't exist, it's empty.
    if not os.path.exists(db_path):
        return True
    return len(connect(db_path)) == 0


class MiddleContainer(Container):
    pass


typer_app = typer.Typer()


@typer_app.command()
def main(db_path: Annotated[str, typer.Argument(help="Path to the ASE database")]):
    if is_db_empty(db_path):
        error = Panel(
            f"The database [bold]{db_path}[/bold] is empty!",
            border_style=STYLE_ERRORS_PANEL_BORDER,
            title=ERRORS_PANEL_TITLE,
            title_align=ALIGN_ERRORS_PANEL,
        )
        console = _get_rich_console(stderr=True)
        console.print(error)
        raise typer.Exit(code=1)
    app = TEXASE(path=db_path)
    app.run()


if __name__ == "__main__":
    typer_app()



from __future__ import annotations

from typing import Any

from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Center, Horizontal, Vertical
from textual.screen import ModalScreen
from textual.widgets import Footer, Input, Label

from texase.formatting import (
    convert_str_to_other_type,
    convert_value_to_int_float_or_bool,
    correctly_typed_kvp,
    kvp_exception,
)
from texase.validators import kvp_validators_add, kvp_validators_edit


class InputScreen(ModalScreen[Any]):
    """Screen with an Input widget."""

    DEFAULT_CSS = """
        InputScreen {
            align: center middle;
        }

        InputScreen > Vertical {
            height: auto;
            content-align: center middle;
        }

        InputScreen > Vertical > Horizontal {
            height: auto;
        }

        #title-label-input-screen {
            text-align: center;
            width: 50%;
            height: auto;
            border: heavy $warning;
            padding: 2 4;
        }

        #input-input-screen {
            border-left: none;
            width: 1fr;
            padding: 0 1 0 0;
        }

        #key-label-input-screen{
            padding: 0 0 0 1;
            background: $boost;
            color: $text;
            width: auto;
            border: tall $accent;
            border-right: none;
        }
    """

    BINDINGS = [
        Binding("ctrl+g", "cancel", "Cancel"),
    ]

    validators = []

    def __init__(
        self, input_title: str, key: str = "", prefilled_input: str | None = None
    ) -> None:
        self.input_title = input_title
        self.prefilled_input = prefilled_input
        self.key = key
        super().__init__()

    def compose(self) -> ComposeResult:
        with Vertical():
            with Center():
                yield Label(self.input_title, id="title-label-input-screen")
            with Horizontal():
                key_label = self.key
                if key_label != "":
                    key_label += " = "
                yield Label(key_label, id="key-label-input-screen")
                yield Input(
                    value=self.prefilled_input,
                    id="input-input-screen",
                    validators=self.validators,
                    validate_on=["submitted"],
                )
        yield Footer()

    def action_cancel(self) -> None:
        self.dismiss(None)

    def on_input_submitted(self, submitted: Input.Submitted) -> None:
        if (
            submitted.validation_result is not None
            and not submitted.validation_result.is_valid
        ):
            self.app.notify_error(
                "\n".join(submitted.validation_result.failure_descriptions),
                error_title="Invalid input",
            )
            # If not valid input stop bubbling further
            submitted.stop()
            return

        self.further_process_submitted(submitted)

    def further_process_submitted(self, submitted: Input.Submitted) -> None:
        raise NotImplementedError("Subclasses must implement this method.")


class KVPScreen(InputScreen):
    """Screen for adding a key value pair."""

    validators = kvp_validators_add

    def further_process_submitted(self, submitted: Input.Submitted) -> None:
        key, value = correctly_typed_kvp(submitted.value)

        exception_from_kvp = kvp_exception(key, value)
        if exception_from_kvp is not None:
            self.app.notify_error(exception_from_kvp, "ValueError")
            return

        self.dismiss((key, value))


class KVPEditScreen(InputScreen):
    """Screen for editing a key value pair."""

    validators = kvp_validators_edit

    def further_process_submitted(self, submitted: Input.Submitted) -> None:
        value = convert_value_to_int_float_or_bool(submitted.value)

        exception_from_kvp = kvp_exception(self.key, value)
        if exception_from_kvp is not None:
            self.app.notify_error(exception_from_kvp, "ValueError")
            return

        self.dismiss(value)


class DataScreen(InputScreen):
    """Screen for adding data."""

    def further_process_submitted(self, submitted: Input.Submitted) -> None:
        value = submitted.value.strip()
        key, data = value.split("=")
        self.dismiss((key, convert_str_to_other_type(data)))


class DataEditScreen(InputScreen):
    """Screen for editing data."""

    def further_process_submitted(self, submitted: Input.Submitted) -> None:
        data = submitted.value.strip()
        self.dismiss(convert_str_to_other_type(data))



from textual import on
from textual.app import ComposeResult
from textual.containers import Horizontal
from textual.css.query import NoMatches
from textual.widgets import Button, Label


class KeyBox(Horizontal):
    async def add_key(self, key: str) -> None:
        new_key = Key(key, id=f"key-{key}")
        await self.mount(new_key)

    def remove_key(self, key: str) -> None:
        key_button = self.query_one(f"#key-{key}", Key)
        key_button.remove()

    def compose(self) -> ComposeResult:
        yield Label("Other keys:")

    async def populate_keys(self, keys: list[str]) -> None:
        """Populate KeyBox with the given keys.

        Also remove superfluous keys."""
        for key_node in self.query(Key).nodes:
            key_name = str(key_node.id).replace("key-", "")
            if key_name not in keys:
                key_node.remove()
                # self.remove_key(key_name)
        for key in keys:
            # Only add the key if it doesn't already exist
            try:
                self.query_one(f"#key-{key}", Key)
            except NoMatches:
                await self.add_key(key)

    @on(Button.Pressed)
    def add_column_to_table(self, event: Button.Pressed) -> None:
        self.app.add_column_to_table_and_remove_from_keybox(str(event.button.label))


class Key(Button):
    DEFAULT_CSS = """
    Key {
        border: none;
        height: 1;
        min-width: 3;
        padding: 0;
        padding-right: 0;
        padding-left: 0;
    }
    """



from rich.text import Text
from textual.app import ComposeResult
from textual.binding import Binding
from textual.screen import ModalScreen
from textual.widgets import Footer, Label


class YesNoScreen(ModalScreen[bool]):
    """Screen with a question that can be answered yes or no."""

    DEFAULT_CSS = """
        YesNoScreen {
            align: center middle;
        }

        YesNoScreen > Label {
            text-align: center;
            width: 50%;
            height: auto;
            border: heavy $warning;
            padding: 2 4;
        }
"""

    BINDINGS = [
        ("y", "yes", "Yes"),
        ("n", "no", "No"),
        Binding("Y", "yes", "Yes", show=False),
        Binding("N", "no", "No", show=False),
        Binding("ctrl+g", "no", "No", show=False),
    ]

    def __init__(self, question: Text) -> None:
        self.question = question
        super().__init__()

    def compose(self) -> ComposeResult:
        yield Label(self.question)
        yield Footer()

    def action_yes(self) -> None:
        self.dismiss(True)

    def action_no(self) -> None:
        self.dismiss(False)



from textual import on
from textual.app import ComposeResult
from textual.containers import Horizontal
from textual.widgets import Input, Label

from texase.validators import kvp_validators_add, kvp_validators_edit


class EditBox(Horizontal):
    BINDINGS = [
        # Hiding boxes shortcut is already defined in app, just
        # redefining the binding here to make it show in the edit
        # box
        ("ctrl+g", "hide_edit", "Undo/Hide"),
    ]

    input_widget_id = "edit-input"
    label_id = "edit-label"

    def compose(self) -> ComposeResult:
        yield Label("Edit:", id=self.label_id, classes="bottom-labels")
        yield Input(
            id=self.input_widget_id,
            classes="bottom-inputs",
            validators=kvp_validators_edit,
            validate_on=["submitted"],
        )

    def focus(self) -> None:
        self.query_one(f"#{self.input_widget_id}").focus()

    def on_hide(self) -> None:
        self.query_one(f"#{self.input_widget_id}", Input).value = ""

    @on(Input.Submitted)
    def show_invalid_reasons(self, event: Input.Submitted) -> None:
        # Updating the UI to show the reasons why validation failed
        if not event.validation_result.is_valid:
            self.query_one(f"#{self.label_id}").add_class("-invalid")
            self.notify(
                "\n".join(event.validation_result.failure_descriptions),
                title="Invalid input",
                severity="error",
            )
        else:
            self.query_one(f"#{self.label_id}").remove_class("-invalid")


class AddBox(EditBox):
    input_widget_id = "add-input"
    label_id = "add-label"

    def compose(self) -> ComposeResult:
        yield Label("Edit:", id=self.label_id, classes="bottom-labels")
        yield Input(
            id=self.input_widget_id,
            classes="bottom-inputs",
            validate_on=["submitted"],
            validators=kvp_validators_add,
        )



from __future__ import annotations

from pathlib import Path
from typing import Iterable

from ase.io.formats import ioformats
from textual import on
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Container
from textual.screen import ModalScreen
from textual.widgets import DirectoryTree, Footer, Input, Label, Tree


def build_write_exts() -> set[str]:
    ext_list = []
    for format in ioformats.values():
        # We are only interested in seeing files that we can append to
        # else just write the filename yourself
        if not format.single and format.can_append:
            if format.extensions:
                ext_list.extend(format.extensions)
            else:
                ext_list.append(format.name)
    return set([f".{ext}" for ext in ext_list])


ASE_IO_WRITE_EXTS = build_write_exts()


def build_read_extensions_and_globs() -> tuple[set[str], set[str]]:
    ext_list = []
    glob_list = []
    for format in ioformats.values():
        if format.can_read:
            # prefer globs to extensions, this is only based on vasp IOFormat
            if format.globs:
                glob_list.extend(format.globs)
            elif format.extensions:
                ext_list.extend(format.extensions)
            else:
                ext_list.append(format.name)
    return set([f".{ext}" for ext in ext_list]), set(glob_list)


ASE_IO_READ_EXTS, ASE_IO_READ_GLOBS = build_read_extensions_and_globs()


class ASEWriteDirectoryTree(DirectoryTree):
    def filter_paths(self, paths: Iterable[Path]) -> Iterable[Path]:
        return [
            path
            for path in paths
            if (path.suffix in ASE_IO_WRITE_EXTS or path.is_dir())
        ]


class ASEReadDirectoryTree(DirectoryTree):
    BINDINGS = [
        Binding("left", "set_root_up", "Go up", show=False),
        Binding("right", "set_root_down", "Go down", show=False),
    ]

    def filter_paths(self, paths: Iterable[Path]) -> Iterable[Path]:
        allowed_paths = []
        for path in paths:
            if path.name.startswith("."):
                # Don't allow hidden files (on Unix defined as starting with a .)
                continue
            elif path.suffix in ASE_IO_READ_EXTS or path.is_dir():
                allowed_paths.append(path)
        for glob in ASE_IO_READ_GLOBS:
            allowed_paths.extend(path.parent.glob(glob))
        return allowed_paths

    def action_set_root_up(self) -> None:
        """If the root node is selected, set a new root node as the
        parent of the current root node."""
        if self.cursor_node is not None:
            if self.cursor_node.is_root:
                self.path = self.path.parent
            elif self.cursor_node.data.path.is_dir():
                self.select_node(self.cursor_node.parent)

    def action_set_root_down(self) -> None:
        """If a directory is selected, set it as the new root node."""
        if self.cursor_node is not None and self.cursor_node.data.path.is_dir():
            self.path = self.cursor_node.data.path
            self.select_node(self.root)


class FilesIOScreen(ModalScreen):  # no typing here to keep Python3.9 happy
    """Screen with a question that can be answered yes or no."""

    BINDINGS = [
        Binding("ctrl+g", "cancel", "Cancel", show=False),
    ]

    def __init__(self, read: bool = True) -> None:
        self.read = read
        super().__init__()

    def compose(self) -> ComposeResult:
        if self.read:
            yield Container(
                Label("Select a file to read from"),
                ASEReadDirectoryTree(Path(".").resolve()),
                FolderLabel("Current folder"),
                Input(),
            )
        else:
            yield Container(
                Label("Select a file to write to"),
                ASEWriteDirectoryTree(Path(".").resolve()),
                FolderLabel("Current folder"),
                Input(),
            )
        yield Footer()

    def action_cancel(self) -> None:
        self.dismiss(None)

    def on_input_submitted(self, submitted: Input.Submitted) -> None:
        self.dismiss(Path(submitted.value))

    def on_directory_tree_file_selected(
        self, selected: DirectoryTree.FileSelected
    ) -> None:
        self.dismiss(selected.path)

    @on(Tree.NodeHighlighted)
    def set_input(self, event: Tree.NodeHighlighted) -> None:
        # Add a / to the end if it is a directory
        path = event.node.data.path
        str_path = str(path)
        if path.is_dir():
            str_path += "/"
        self.query_one(Input).value = str_path


class FolderLabel(Label): ...



from __future__ import annotations

import ast
import re
from typing import Any, Callable, Tuple

import numpy as np
import pandas as pd
from ase.db.core import check, float_to_time_string, now
from rich.text import Text

# The labels showing marked and unmarked rows
MARKED_LABEL = Text("\u25cf", style="bright_yellow")
UNMARKED_LABEL = Text("\u2219", style="grey")


def format_value(val) -> Text | str:
    if pd.isna(val):
        return ""
    if isinstance(val, str):
        return val
    elif isinstance(val, float):  # and not val.is_integer():
        if abs(val) > 1e6 or abs(val) < 1e-3:
            format_spec = "#.3g"
        else:
            format_spec = ".2f"
        return Text("{1:{0}}".format(format_spec, val), justify="right")
    # Checking for integers, see this helpful diagram:
    # https://numpy.org/doc/stable/reference/arrays.scalars.html
    # And this answer: https://stackoverflow.com/a/37727662
    elif np.issubdtype(type(val), np.integer):
        return Text(str(val), justify="right")
    else:
        return str(val)


# Mapping from data.df column dtype to the output dtype used in the
# table. This dict is used when going from data.df to the table, so
# the keys in the dict should be all the possible dtypes in data.df.
series_dtype_to_output_dtype = {
    pd.Int64Dtype(): "object",
    pd.BooleanDtype(): pd.StringDtype(),
    pd.StringDtype(): pd.StringDtype(),
    np.dtype("int"): "object",
    np.dtype("float"): "object",
    np.dtype("object"): "object",
    "int": "object",
    "object": "object",
    "float": "object",
}


def format_column(
    col: pd.Series, format_function: Callable = format_value
) -> pd.Series:
    return pd.Series(
        [format_function(val) for val in col],
        dtype=series_dtype_to_output_dtype[col.dtype],
        name=col.name,
    )


def get_age_string(ctime) -> str:
    return float_to_time_string(now() - ctime)


def convert_value_to_int_float_or_bool(value):
    """Convert value to int, float or bool if possible. Otherwise return the value as is.

    Modified from ASE.
    """
    try:
        return int(value)
    except ValueError:
        try:
            value = float(value)
        except ValueError:
            value = {"true": True, "false": False}.get(value.lower(), value)
        return value


def convert_str_to_other_type(str_value: str) -> Any:
    """Convert string to dict, list or np.ndarray if possible. Otherwise return the value as is."""
    try:
        return ast.literal_eval(str_value)
    except (SyntaxError, ValueError):
        # Maybe this is a numpy array
        if is_numpy_array(str_value):
            return np.array(string_to_list(str_value))
        return str_value


def string_to_list(input_string):
    # Remove np.array if present
    if "np.array" in input_string:
        input_string = re.search(r"\[(.*)\]", input_string).group(0)

    # Replace spaces with commas for a proper list format if necessary
    if " " in input_string and "," not in input_string:
        input_string = input_string.replace(" ", ", ")

    # Convert string to list
    return ast.literal_eval(input_string)


def is_numpy_array(s):
    """Match with a regexp one of these two patterns:

    '[1 2 3 ...]'
    and
    'np.array([1, 2, 3 ...])'

    Thanks to copilot for the pattern.
    """
    # Regular expression pattern to match arrays of exclusively numbers or strings
    number_array_pattern = r"\[\d+(\.\d+)?( \d+(\.\d+)?)*\]"
    string_array_pattern = r'\["[^"]*"( "[^"]*")*\]'
    numpy_number_array_pattern = r"np\.array\(\[\d+(\.\d+)?(, \d+(\.\d+)?)*\]\)"
    numpy_string_array_pattern = r'np\.array\(\["[^"]*"(, "[^"]*")*\]\)'

    # Combine the patterns to match either numbers or strings but not both
    combined_pattern = r"^({}|{}|{}|{})$".format(
        number_array_pattern,
        string_array_pattern,
        numpy_number_array_pattern,
        numpy_string_array_pattern,
    )

    # Search for the pattern in the string
    return bool(re.match(combined_pattern, s))


def pbc_str_to_array(pbc_str: str) -> np.ndarray:
    """Convert e.g. the string TFT to [True, False, True]"""
    return np.array([c == "T" for c in pbc_str.upper()])


def kvp_exception(key, value) -> str | None:
    """Check that key-value-pair is valid for ase.db

    It is ok to edit pbc, we make this check first."""

    if key == "pbc":
        try:
            check_pbc_string_validity(value)
        except ValueError as e:
            return str(e)
        return None

    try:
        check({key: value})
    except ValueError as e:
        # Notify that the key-value-pair is not valid with the
        # raised ValueError and then return
        return str(e)
    return None


def check_pbc_string_validity(string):
    # check if the string has exactly three characters
    if len(string) == 3:
        # convert the string to upper case
        string = string.upper()
        # loop through each character in the string
        for char in string:
            # check if the character is either t or f
            if char not in ["T", "F"]:
                # raise a ValueError with a descriptive message
                raise ValueError(f"{string} contains characters that are not T or F!")
        # return True if all characters are t or f
        return True
    else:
        # raise a ValueError with a descriptive message
        raise ValueError(f"{string} does not have exactly three characters!")


def correctly_typed_kvp(input_str: str) -> Tuple[str, Any]:
    """Convert the input string of the form 'key=value' to the correct
    type if possible. At this point the input should be validated,
    i.e. the value contains a = and the key/column is editable

    """
    # Split the input string into a key and a value
    key, value = input_str.split("=", 1)
    # Remove leading and trailing whitespace
    key = key.strip()
    value = value.strip()
    # Convert the value to the correct type if possible
    value = convert_value_to_int_float_or_bool(value)

    return key, value



import json
import os
from pathlib import Path


class SavedColumns:
    def __init__(self):
        self._columns_file_path = Path(
            os.environ.get("TEXASE_COLUMNS_FILE", Path.home() / ".texase-columns.json")
        )
        self._files_and_columns = self._read_columns_file()

    def _read_columns_file(self):
        if self._columns_file_path.exists():
            with self._columns_file_path.open("r") as f:
                return json.load(f)
        return {}

    def _write_columns_file(self):
        with self._columns_file_path.open("w") as f:
            json.dump(self._files_and_columns, f, indent=4)

    def __getitem__(self, key):
        return self._files_and_columns.get(key, None)

    def __setitem__(self, key, value):
        self._files_and_columns[key] = value
        self._write_columns_file()

    def __delitem__(self, key):
        del self._files_and_columns[key]
        self._write_columns_file()

    def __len__(self):
        return len(self._files_and_columns)

    def __str__(self):
        return str(self._files_and_columns)

    def __repr__(self):
        return repr(self._files_and_columns)



from __future__ import annotations

from typing import Iterable, List, Set, Tuple, Union

from rich.text import Text
from textual.binding import Binding
from textual.coordinate import Coordinate
from textual.widgets import DataTable, Input, Label
from textual.widgets._data_table import ColumnKey, RowKey

from texase.data import ALL_COLUMNS, Data
from texase.edit import AddBox, EditBox
from texase.formatting import MARKED_LABEL, UNMARKED_LABEL, format_value

UNEDITABLE_COLUMNS = [c for c in ALL_COLUMNS if c not in ["pbc"]]


class TexaseTable(DataTable):
    BINDINGS = [
        ("q", "quit", "Quit"),
        ("?", "toggle_help", "Help"),
        ("s", "sort_column", "Sort"),
        ("space", "mark_row", "Mark row"),
        Binding("enter", "toggle_details", "Show details", show=False),
        ("e", "edit", "Edit"),
        ("k", "add_key_value_pair", "Add key-value pair"),
        ("#", "delete_rows", "Delete row(s)"),
        ("d", "delete_key_value_pairs", "Delete key-value pair(s)"),
        ("x", "export_rows", "Export row(s)"),
        ("i", "import_rows", "Import structure(s)"),
        ("v", "view", "View"),
        Binding("+", "add_column", "Add column", show=False),
        Binding("-", "remove_column", "Hide column", show=False),
        Binding("u", "unmark_row", "Unmark row", show=False),
        Binding("U", "unmark_all", "Unmark all", show=False),
        Binding("ctrl+s", "search", "Search"),
        Binding("/", "filter", "Filter rows"),
        Binding("g", "update_view", "Update from database", show=False),
    ]

    marked_rows: Set = set()

    def _manipulate_filters(
        self, filter_tuple: Tuple[str, str, str], add: bool = True
    ) -> None:
        # Clear the table, but not the columns
        self.clear()

        # Get the key, operator and value from the tuple
        key, operator, value = filter_tuple

        if add:
            # Add the filter to the data object
            self.ancestors[-1].data.add_filter(key, operator, value)
        else:
            # Remove the filter from the data object
            self.ancestors[-1].data.remove_filter(filter_tuple)

        # Rebuild the table
        self.populate_table(
            self.ancestors[-1].data, marked_rows=self.marked_rows, columns_cleared=False
        )

    def add_filter(self, key, operator, value) -> None:
        self._manipulate_filters((key, operator, value), add=True)

    def remove_filter(self, key, operator, value) -> None:
        self._manipulate_filters((key, operator, value), add=False)

    def populate_table(
        self,
        data: Data,
        marked_rows: Union[List[int], None] = None,
        *,
        columns_cleared: bool = True,
    ) -> None:
        # Columns
        if columns_cleared:
            for col in data.chosen_columns:
                self.add_column(col, key=col)

        # Get ready for handling marked rows
        marked_row_keys = set()

        # Populate rows by fetching data
        for row in data.df_for_print().itertuples(index=False):
            if marked_rows is not None and RowKey(str(row[0])) in marked_rows:
                row_key = self.add_row(*row, key=str(row[0]), label=MARKED_LABEL)
                marked_row_keys.add(row_key)
            else:
                row_key = self.add_row(*row, key=str(row[0]), label=UNMARKED_LABEL)
        self.marked_rows = marked_row_keys

    def add_table_rows(self, data: Data, indices: Iterable[int]) -> None:
        for row in data.df_for_print().iloc[indices].itertuples(index=False):
            self.add_row(*row, key=str(row[0]), label=UNMARKED_LABEL)

    def update_table_rows(self, data: Data, indices: Iterable[int]) -> None:
        for index in indices:
            row = data.df_for_print().iloc[index]
            # Iterate through the columns of row
            for col in data.chosen_columns:
                self.update_cell(RowKey(str(row.id)), ColumnKey(col), row[col])

    def check_columns(self, data: Data) -> None:
        """Check if the columns shown in the table are up-to-date with
        data.chosen_columns, if not remove from the table."""
        for col in list(self.columns.keys()):
            if col not in data.chosen_columns:
                self.remove_column(col)

    def add_column_and_values(self, column_name: str) -> None:
        """Add a column and its values to the table.

        It is assumed that the column is present in the data.

        Parameters
        ----------
        column_name : str
            The name of the column to add.
        """
        col_key = self.add_column(column_name, key=column_name)
        col_index = self.get_column_index(col_key)

        # Column_for_print gets the values in the same order
        # as shown in the table, thus we can just use
        # enumerate to get the row index
        values = self.app.data.column_for_print(column_name)
        for i, val in enumerate(values[:-1]):
            self.update_cell_at(Coordinate(i, col_index), val)

        # Clear _updated_cells to avoid a forced update for each cell
        # in the function _update_column_widths
        self._updated_cells.clear()

        self.update_cell_at(
            Coordinate(len(values) - 1, col_index),
            values.iloc[-1],
            update_width=True,
        )

    def is_cell_editable(
        self, uneditable_columns: List[str] = UNEDITABLE_COLUMNS
    ) -> bool:
        # Check if current cell is editable
        coordinate = self.cursor_coordinate

        # Get current column name
        column_name = str(list(self.columns.values())[coordinate.column].label)
        if column_name in uneditable_columns:
            self.notify(
                f"Column {column_name} can not be edited!",
                severity="warning",
                title="Warning",
            )
            return False
        else:
            return True

    def update_row_editable_cells(self, key_value_pairs: dict) -> None:
        """Update the editable cells of a row.

        This should be called from details, thus it is assumed that
        the cursor position hasn't changed and we can just use the
        cursor row.

        Parameters
        ----------
        row_id : int
            The row id of the row to update.
        key_value_pairs : dict
            A dictionary of key-value pairs to update.

        """
        for key, value in key_value_pairs.items():
            # Check that the column is actually shown
            if key not in self.columns:
                continue

            # Get the column index
            col_index = self.get_column_index(key)

            # Update the cell
            self.update_cell_at(
                Coordinate(self.cursor_row, col_index),
                format_value(value),
                update_width=True,
            )

    def update_edit_box(self, editbox: EditBox) -> None:
        # Current cell should be editable
        coordinate = self.cursor_coordinate

        # Get current column name
        column_name = self.column_at_cursor()

        label_str = f"Edit [bold]{column_name} =[/bold]"
        editbox.query_one("#edit-label", Label).renderable = label_str

        # Get current cell value
        cell_value = self.get_cell_at(coordinate)
        editbox.query_one("#edit-input", Input).value = str(cell_value)

        # Special rules to edit e.g. pbc or volume (then get cell editor) etc.

    def update_cell_from_edit_box(self, new_value: Text | str) -> None:
        self.update_cell_at(self.cursor_coordinate, new_value, update_width=True)

    def column_at_cursor(self) -> str:
        """Return the name of the column at the cursor."""
        return str(list(self.columns.values())[self.cursor_column].label)

    # Add/Edit/Delete key value pairs
    def update_add_box(self, addbox: AddBox) -> None:
        # Set the value of the key input field to an empty string
        input_field = addbox.query_one("#add-input", Input)
        input_field.value = ""
        input_field.placeholder = "key = value"

        # Update label to show add kvp
        label_str = "Add/Edit key value pair:"
        addbox.query_one("#add-label", Label).renderable = label_str

    def update_cell_from_add_box(self, column: str, value: Text | str) -> None:
        # Is column already present in the table? If so update the cells
        if column in self.columns:
            column_key = ColumnKey(column)
            if len(self.marked_rows) > 0:
                marked_list = list(self.marked_rows)
                for row_key in marked_list[:-1]:
                    self.update_cell(row_key, column_key, value)
                self.update_cell(marked_list[-1], column_key, value, update_width=True)
            else:
                # Add the column to the current row
                self.update_cell_at(
                    Coordinate(self.cursor_row, self.get_column_index(column_key)),
                    value,
                    update_width=True,
                )
        # If not, add the column. The values is already set in the data.
        else:
            self.add_column_and_values(column)

    def delete_kvp_question(self) -> Text:
        """Return the question to ask when deleting a key value pair."""
        # Get the ids of the marked or currently selected rows
        ids = self.ids_to_act_on()
        no_rows = len(ids)

        # Get the column name
        column_name = self.column_at_cursor()

        # Create the question
        ids_str = ", ".join([str(id) for id in ids])
        plural = [" ", "s "][no_rows > 1]
        q = "Do you want to delete the key value pair" + plural
        q += f"[bold]{column_name}[/bold] in row" + plural
        q += "with id" + plural + f"[bold]{ids_str}[/bold]?"
        return Text.from_markup(q)

    def delete_selected_key_value_pairs(self) -> None:
        """Delete key value pairs of the currently highlighted
        column. If some rows are marked then delete the key value
        pairs of all marked rows, else delete the key value pair of
        the current row."""

        rows = self.row_keys_to_act_on()

        for row_key in rows:
            self.update_cell(
                row_key, self.column_index_to_column_key(self.cursor_column), ""
            )

    def delete_row_question(self) -> Text:
        """Return the question to ask when deleting a key value pair."""
        # Get the ids of the marked or currently selected rows
        ids = self.ids_to_act_on()
        no_rows = len(ids)

        # Create the question
        ids_str = ", ".join([str(id) for id in ids])
        plural = [" ", "s "][no_rows > 1]
        q = "Do you want to delete the row" + plural
        q += "with id" + plural + f"[bold]{ids_str}[/bold]?"
        return Text.from_markup(q)

    def delete_selected_rows(self) -> None:
        """Delete the currently selected rows from the table view."""
        self.delete_rows(self.row_keys_to_act_on())

    def delete_rows(self, row_keys: Iterable[RowKey]) -> None:
        for row_key in row_keys:
            self.remove_row(row_key)
            self.marked_rows.discard(row_key)

    # Selecting/marking rows
    def action_mark_row(self) -> None:
        row_key = self.row_index_to_row_key(self.cursor_row)
        self.toggle_mark_row(row_key)

        # Go to the next row after marking
        self.action_cursor_down()

        # # Best effort so far for highlighting the background of a row
        # marked_text = Text(style=table.get_component_styles("datatable--cursor").rich_style)
        # for column_index, cell in enumerate(table.get_row_at(row)):
        #     if isinstance(cell, Text):
        #         cell.style = table.get_component_styles("datatable--cursor").rich_style
        #         table.update_cell_at(Coordinate(row, column_index), cell)
        #     else:
        #         table.update_cell_at(Coordinate(row, column_index), marked_text + cell)

    def action_unmark_row(self) -> None:
        row_index = self.cursor_row
        row_key = list(self.rows.keys())[row_index]
        self.unmark_row(row_key)

        # Go to the next row after unmarking
        self.action_cursor_down()

    def action_unmark_all(self) -> None:
        # Remove all marked rows in one go, this requires a full table
        # refresh. I don't know if it would be faster to call
        # action_unmark_row() for each row.
        for row_key in self.marked_rows:
            self.rows[row_key].label = UNMARKED_LABEL
        self.marked_rows = set()
        self._update_count += 1
        self.refresh()

    def on_data_table_row_label_selected(
        self, selected: DataTable.RowLabelSelected
    ) -> None:
        self.toggle_mark_row(selected.row_key)
        self.update_row_after_mark_operation(row_index=selected.row_index)

    def toggle_mark_row(self, row_key: RowKey) -> None:
        if row_key in self.marked_rows:
            self.unmark_row(row_key)
        else:
            self.mark_row(row_key)

    def unmark_row(self, row_key: RowKey) -> None:
        """Unmark a row.

        If it's not marked, do nothing.
        """
        if row_key in self.marked_rows:
            self.marked_rows.remove(row_key)
            self.rows[row_key].label = UNMARKED_LABEL
            self.update_row_after_mark_operation(row_key=row_key)

    def mark_row(self, row_key: RowKey) -> None:
        """Mark a row.

        If it is already marked, then it stays marked.
        """
        self.rows[row_key].label = MARKED_LABEL
        self.marked_rows.add(row_key)

        self.update_row_after_mark_operation(row_key=row_key)

    def update_row_after_mark_operation(
        self, /, row_key: Union[RowKey, None] = None, row_index: Union[int, None] = None
    ) -> None:
        """Tell the table that something has changed so it will refresh properly"""
        self._update_count += 1
        if row_index is not None:
            self.refresh_row(row_index)
        elif row_key is not None:
            self.refresh_row(self.get_row_index(row_key))

    def get_marked_row_ids(self) -> List[int]:
        """Return the ids of the rows that are currently marked"""
        return [get_id_from_row(self.get_row(row_key)) for row_key in self.marked_rows]

    def row_id_at_cursor(self) -> int:
        """Return the row id at the cursor as an int."""
        return get_id_from_row(self.get_row_at(self.cursor_row))

    def ids_to_act_on(self) -> List[int]:
        """Get the ids of the rows to act on using the same logic as
        row_keys_to_act_on."""
        return sorted(
            [
                get_id_from_row(self.get_row(row_key))
                for row_key in self.row_keys_to_act_on()
            ]
        )

    def row_keys_to_act_on(self) -> List[RowKey]:
        """Get the row keys of the rows to act on. If some rows are marked,
        then return the row keys of those rows, otherwise return a list of the RowKey of
        the row where the cursor is."""
        if self.marked_rows:
            rows = list(self.marked_rows)
        else:
            rows = [self.row_index_to_row_key(self.cursor_row)]
        return rows

    def row_index_to_row_key(self, row_index) -> RowKey:
        """Return the row key of the row at the given row index"""
        return self.coordinate_to_cell_key(Coordinate(row_index, 0)).row_key

    def column_index_to_column_key(self, column_index) -> ColumnKey:
        """Return the column key of the column at the given column index"""
        return self.coordinate_to_cell_key(Coordinate(0, column_index)).column_key


def get_id_from_row(row) -> int:
    # This assumes that the first index of the row is the id
    return int(str(row[0]))


def get_column_labels(columns) -> list:
    return [str(c.label) for c in columns.values()]



from __future__ import annotations

import operator
from collections import defaultdict
from dataclasses import dataclass, field
from datetime import datetime
from functools import wraps
from itertools import combinations
from pathlib import Path
from typing import Any, Dict, Iterable, List, Tuple, Union, overload

import numpy as np
import pandas as pd
from ase import Atoms
from ase.db import connect
from ase.db.table import all_columns
from ase.io import read, write
from textual.cache import LRUCache

from texase.formatting import format_column, get_age_string, pbc_str_to_array
from texase.saved_columns import SavedColumns

ops = {
    "==": operator.eq,
    "!=": operator.ne,
    "<": operator.lt,
    ">": operator.gt,
    "<=": operator.le,
    ">=": operator.ge,
}


# If the operator is the key then convert the comparison value with
# the function specified in the value
def nothing(x):
    return x


ALL_COLUMNS = list(all_columns) + ["modified"]
COLUMN_DTYPES = {
    "id": "int",
    "age": "float",
    "user": pd.StringDtype(),
    "formula": pd.StringDtype(),
    "calculator": pd.StringDtype(),
    "energy": "float",
    "natoms": "int",
    "fmax": "float",
    "pbc": pd.StringDtype(),
    "volume": "float",
    "charge": "float",
    "mass": "float",
    "smax": "float",
    "magmom": "float",
}


def get_default_columns():
    """Return default columns used in ASE db and here, i.e. don't show
    modified by default."""
    return list(all_columns)


operator_type_conversion = {
    "<": float,
    "<=": float,
    "==": nothing,
    ">=": float,
    ">": float,
    "!=": nothing,
}


def cache(f):
    @wraps(f)
    def wrapper(self, *args, **kwds):
        cache_key = tuple(args)
        cache = getattr(self, f.__name__ + "_cache")
        result = cache.get(cache_key, None)
        if result is not None:
            return result
        result = f(self, *args, **kwds)
        cache[cache_key] = result
        return result

    return wrapper


class ASEReadError(Exception):
    """Error when ASE tries to read a file."""


class ASEWriteError(Exception):
    """Error when ASE tries to write to a file."""


@dataclass
class Data:
    df: pd.DataFrame
    db_path: Path
    user_keys: List[str]
    chosen_columns: list = field(default_factory=get_default_columns)
    saved_columns: Union[SavedColumns, None] = None
    data_filter: Union[List[Tuple[str, str, str]], None] = None
    sort_reverse: bool = False

    def __post_init__(self):
        self._filters = tuple()
        if self.data_filter is not None:
            self.add_filter(*self.data_filter)
        self.db_path = Path(self.db_path).resolve()
        self.saved_columns = SavedColumns()
        self.update_chosen_columns()
        self.sort_columns: List[str] = ["id"]
        self._filter_mask_cache: LRUCache[tuple, np.ndarray] = LRUCache(maxsize=128)
        self._string_df_cache: LRUCache[tuple, pd.DataFrame] = LRUCache(maxsize=128)
        self._string_column_cache: LRUCache[tuple, pd.Series] = LRUCache(maxsize=128)
        self.last_update_time = datetime.now()

    def unused_columns(self) -> List[str]:
        """Returns a list of columns that are not used in the table.

        Go through all columns + self.user_keys and check if they are
        in self.chosen_columns

        """
        return [
            col
            for col in ALL_COLUMNS + self.user_keys
            if col not in self.chosen_columns
        ]

    @overload
    def update_value(
        self, ids: int, column: str, value: str | float | int | None
    ) -> None: ...

    @overload
    def update_value(
        self, ids: Iterable[int], column: str, value: str | float | int | None
    ) -> None: ...

    def update_value(self, ids, column, value) -> None:
        """Updates the value in the database and self.df

        ids: list of ids
        column: column name
        value: new value
        """
        # Test if ids is iterable
        try:
            iter(ids)
        except TypeError:
            ids = [ids]

        if value is None:
            self.update_in_db(ids, delete_keys=[column])
        else:
            self.update_in_db(ids, {column: value})

        # Update in self.df
        for row_id in ids:
            self.update_df(row_id, column, value)

        # Clear the caches
        self._string_df_cache.clear()
        self._remove_edited_column_from_caches(column)

    def is_column_empty(self, column: str) -> bool:
        """Check if the column is empty"""
        return self.df[column].isna().all()

    def clear_all_caches(self) -> None:
        """Clear all caches"""
        self._filter_mask_cache.clear()
        self._string_df_cache.clear()
        self._string_column_cache.clear()

    def update_in_db(
        self,
        row_ids: int | Iterable[int],
        key_value_pairs: Dict[str, Any] = {},
        data: dict | None = None,
        delete_keys: list = [],
    ) -> None:
        """Update the row id(s) db with the given key value pairs and data"""
        try:
            iter(row_ids)  # type: ignore
        except TypeError:
            row_ids = [row_ids]  # type: ignore

        atoms = None
        with connect(self.db_path) as db:
            for idx in row_ids:
                atoms = None
                if key_value_pairs:
                    if "pbc" in key_value_pairs:
                        # We need to remove pbc from key_value_pairs,
                        # thus first we copy the dict and then pop pbc
                        # from the copy
                        key_value_pairs = dict(key_value_pairs)

                        # Get the atoms and modify directly
                        atoms = db.get_atoms(idx)
                        atoms.pbc = pbc_str_to_array(key_value_pairs.pop("pbc"))
                db.update(
                    idx, atoms, **key_value_pairs, delete_keys=delete_keys, data=data
                )

    def update_row(
        self, row_id: int, key_value_pairs: dict = {}, data: dict | None = None
    ) -> None:
        """Update the row id db with the given key value pairs and data.

        It is assumed that only editable keys are passed in key_value_pairs."""

        delete_keys = []
        for key, value in key_value_pairs.items():
            self.update_df(row_id, key, value)
            # Create delete_keys with the keys in key_value_pairs that have None as value
            if value is None:
                delete_keys.append(key)
        # Remove the keys in key_value_pairs that are going to be deleted
        for key in delete_keys:
            key_value_pairs.pop(key)

        self.update_in_db(row_id, key_value_pairs, data, delete_keys=delete_keys)

    def update_df(
        self, row_id: int, column: str, value: str | float | int | None
    ) -> None:
        idx = self.index_from_row_id(row_id)
        if column in self.df.columns:
            orig_dtype = self.df[column].dtype

            # Check if type of value is compatible with the dtype of
            # the column, if not convert the column dtype to object
            if not is_dtype_compatible_with_value(self.df[column].dtype, value):
                self.df[column] = self.df[column].astype(object)
            self.df.loc[idx, column] = value

            # A value has changed this means that the dtype of the
            # column could be better represented by something other
            # than object, i.e. only applicable if the original dtype
            # was object.
            # TODO this could be slow -> Test
            if orig_dtype == np.dtype("object"):
                self.df[column] = self.df[column].astype(
                    recommend_dtype(self.df[column])
                )
        else:
            new_col = [None] * len(self.df)
            new_col[idx] = value
            dtype = recommend_dtype(new_col)
            new_col = pd.Series(new_col, name=column, dtype=dtype)
            self.df = pd.concat([self.df, new_col], axis=1)

    def export_rows(self, row_ids: Iterable[int], path: Path) -> None:
        with connect(self.db_path) as db:
            append = False
            if path.is_file():
                # An existing file, we append to it
                append = True
            try:
                write(
                    path,
                    [
                        db.get_atoms(idx, add_additional_information=True)
                        for idx in row_ids
                    ],
                    append=append,
                )
            except Exception as e:
                raise ASEWriteError(repr(e))

    def import_rows(self, path: Path, index=-1) -> Iterable[int]:
        """Import atoms from a file and add them to the database and the df.

        Default is to only take the last frame in the file. Change this with the index argument."""
        try:
            atoms_list = read(path, index=index)
        except Exception as e:
            raise ASEReadError(repr(e))

        if isinstance(atoms_list, Atoms):
            atoms_list = [atoms_list]
        with connect(self.db_path) as db:
            new_rows = []
            for atoms in atoms_list:
                new_row = db.write(
                    atoms,
                    key_value_pairs=atoms.info.get("key_value_pairs", {}),
                    data=atoms.info.get("data", {}),
                )
                new_rows.append(new_row)

        return self.add_rows_to_df(sel=f"id>={new_rows[0]}")

    def add_rows_to_df(self, sel="") -> np.ndarray:
        """Get the new rows from the database and add them to
        self.df. We have to get them from the database to get the id
        and ctime correctly.

        Returns the indices of the added rows in the df.
        """
        new_df, new_user_keys = db_to_df(connect(self.db_path), sel=sel)
        original_last_index = self.df.index[-1]

        # Check that the dtypes of common columns are compatible
        for col in set(new_user_keys) & set(self.user_keys):
            if not is_dtypes_compatible(self.df[col].dtype, new_df[col].dtype):
                # Not compatible, convert the column to object
                self.df[col] = self.df[col].astype(object)
        self.df = pd.concat([self.df, new_df], ignore_index=True)

        new_last_index = self.df.index[-1]
        added_indices = np.arange(original_last_index + 1, new_last_index + 1)

        # Add the new user keys to the KeyBox and the unused columns
        self.user_keys = list(set(self.user_keys) | set(new_user_keys))

        # Clear the caches, this is simpler than updating them with
        # the new rows.
        self.clear_all_caches()

        return added_indices

    def add_remaining_rows_to_df(self) -> np.ndarray:
        """After the initial load of rows from the db (with a
        limit). This function checks if there are more rows in the db
        and adds them to the dataframe."""

        # Get the number of rows in the db
        with connect(self.db_path) as db:
            n_rows = db.count()

        # If there are more rows than the initial load, we add them
        if n_rows > len(self.df):
            return self.add_rows_to_df(sel=f"id>{self.df.id.iloc[-1]}")
        return []

    def index_from_row_id(self, row_id) -> int:
        return self.df.loc[self.df["id"] == row_id].index[0]

    def _remove_edited_column_from_caches(self, column) -> None:
        self._string_column_cache.discard((column,))
        for key in self._filter_mask_cache.keys():
            if column in key:
                self._filter_mask_cache.discard(key)

    def df_for_print(self) -> pd.DataFrame:
        """Returns the final dataframe after applying all filters and current sort."""
        df = self.string_df()
        return apply_filter_and_sort_on_df(df, self.filter_mask, self._sort)

    def column_for_print(self, column):
        """Get a string representation of a column in the DataFrame
        including filters and sorting."""
        col = self._string_column(column)
        return col.iloc[self._sort].iloc[self.filter_mask[self._sort]]

    def row_details(self, row_id: int) -> Tuple[dict, dict]:
        """Returns key value pairs from the row in two dictionaries:

        Static and editable key value pairs.
        All user keys and pbc are editable.
        """
        static_kvps = {}
        dynamic_kvps = {}
        editable_keys = self.user_keys + ["pbc"]
        for key, value in (
            self.df.loc[self.df["id"] == row_id].squeeze().dropna().items()
        ):
            if key in editable_keys:
                dynamic_kvps[key] = value
            else:
                static_kvps[key] = value
        return static_kvps, dynamic_kvps

    def row_data(self, row_id: int) -> dict:
        return get_data(self.db_path, row_id)

    def get_atoms(self, row) -> Atoms:
        db = connect(self.db_path)
        return db.get_atoms(id=row)

    def can_column_be_added(self, column) -> bool:
        """Check if a column can be added to the table, i.e. is it
        present in the data but not in the table."""
        return column in self.unused_columns()

    def add_to_chosen_columns(self, column) -> None:
        """Add a column to the table.

        Before calling this function it is checked that the column is valid.
        """
        self.chosen_columns.append(column)

    def remove_from_chosen_columns(self, column) -> bool:
        # Check that the column is in chosen_columns
        if column in self.chosen_columns:
            self.chosen_columns.remove(column)
            return True
        # Return False if nothing has changed
        return False

    def update_chosen_columns(self) -> None:
        # Check if db columns file exists. If so set chosen columns based on that.
        saved_columns = self.saved_columns[str(self.db_path)]
        if saved_columns is not None:
            self.chosen_columns = saved_columns

    def save_chosen_columns(self) -> None:
        self.saved_columns[str(self.db_path)] = self.chosen_columns

    def search_for_string(self, search_string: str, regex: bool = True):
        # Use the string representation of the dataframe, i.e. what is
        # currently visible
        df = self.df_for_print()
        mask = np.column_stack(
            [
                df[col].astype(str).str.contains(search_string, na=False, regex=regex)
                for col in self.chosen_columns
            ]
        )
        return zip(*mask.nonzero())

    @property
    def _sort(self) -> np.ndarray:
        return self.df.sort_values(
            self.sort_columns, ascending=not self.sort_reverse
        ).index.to_numpy()

    def sort(self, col_name: str) -> np.ndarray:
        """Set the indices to sort self.df in self._sort. Return the sorted ids."""

        # Determine if the sort should be reversed
        if len(self.sort_columns) > 0 and col_name == self.sort_columns[0]:
            # If the column is already the first in the sort order, toggle the sort order
            self.sort_reverse = not self.sort_reverse
        else:
            # Otherwise, add/move the column to the sort order at first
            # position and set the sort order to ascending
            if col_name in self.sort_columns:
                self.sort_columns.remove(col_name)
            self.sort_columns.insert(0, col_name)
            self.sort_reverse = False

        return self.id_array_with_filter_and_sort()

    def id_array_with_filter_and_sort(
        self, filter_mask: np.ndarray | None = None
    ) -> np.ndarray:
        """Returns an array of ids after applying filters and sorting."""
        if filter_mask is None:
            filter_mask = self.filter_mask
        return apply_filter_and_sort_on_df(self.df, filter_mask, self._sort)[
            "id"
        ].to_numpy()

    @property
    def filter_mask(self) -> np.ndarray:
        """Combine all boolean arrays from the filters into one."""
        mask = np.ones(len(self.df), dtype=bool)
        for filter in self._filters:
            mask &= self._filter_mask(filter)
        return mask

    @cache
    def _filter_mask(self, filter: tuple) -> np.ndarray:
        """Returns a boolean array of indices that pass the filter"""
        filter_key, op, filter_value = filter
        mask = ops[op](self.df[filter_key], operator_type_conversion[op](filter_value))
        return mask.to_numpy()

    def add_filter(self, key, operator, value) -> None:
        # We get the value as a string. Maybe we should convert it to
        # the correct type if the column values are not strings? But
        # how do we know? We try to deduce from the operator

        self._filters += ((key, operator, value),)

    def remove_filter(self, filter_tuple: Tuple[str, str, str]) -> None:
        self._filters = tuple(
            filter for filter in self._filters if filter != filter_tuple
        )

    def string_df(self) -> pd.DataFrame:
        return self._string_df(tuple(self.chosen_columns))

    @cache
    def _string_df(self, chosen_columns) -> pd.DataFrame:
        """Get a representation of the DataFrame where all values are strings.

        This is used for displaying the data in the table. The
        function depends on self.chosen_columns. The result is cached
        in self._string_df_cache. Thus the cache key will be built
        with self.chosen_columns.

        """
        df_list = []
        for column in chosen_columns:
            df_list.append(self._string_column(column))
        return pd.concat(df_list, axis=1).fillna("", axis=1)

    def get_mask_of_df_with_filter(
        self, filter_tuple: Tuple[str, str, str]
    ) -> np.ndarray:
        return self._filter_mask(filter_tuple)

    @cache
    def _string_column(self, column: str) -> pd.Series:
        """Get a string representation of a column in the raw
        DataFrame without filters or sorting."""
        df = self.df
        column_data = df[column]
        if column in ["age", "modified"]:
            column_data = format_column(column_data, format_function=get_age_string)
        return format_column(column_data)

    def db_last_modified(self) -> datetime:
        file_stat = self.db_path.stat()
        # Get the modification time as a datetime object
        return datetime.fromtimestamp(file_stat.st_mtime)

    def is_df_up_to_date(self) -> bool:
        return self.db_last_modified() <= self.last_update_time

    def updates_from_db(self) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Update df and table from the db.

        Only do anything if the db has been modified since last time
        df and table were updated.

        This is called from the app so it should return
        information about which rows to delete in the table, which
        rows to update in the table and which rows to insert in the
        table.
        """
        if not self.is_df_up_to_date():
            self.last_update_time = datetime.now()

            # Read the columns id and mtime from the db to see which
            # rows have been modified compared to the current df
            with connect(self.db_path) as db:
                update_cache = True
                ids, mtimes = ids_and_mtimes(db)

                # First get the rows to delete. This will influence
                # the indices in the df the rows to add or update.
                # Get the df index of the rows that have been deleted in the db
                indices_to_remove = np.nonzero(~self.df.id.isin(ids))[0]
                if len(indices_to_remove) > 0:
                    self.delete_rows_from_df(
                        indices_to_remove, update_cache=update_cache
                    )

                # Read rows from db with id larger than self.df.id[-1]
                # and add them to self.df
                if np.any(ids > self.df.id.iloc[-1]):
                    indices_to_add = self.add_rows_to_df(f"id>{self.df.id.iloc[-1]}")
                    # All caches are cleared now no need to update the
                    # cache manually in any of the following functions
                    update_cache = False
                else:
                    indices_to_add = np.array([])

                # Get the df index of the rows that have been modified in the db
                # mtimes and self.df.modified should have the same length now.
                indices_to_update = np.nonzero(self.df.modified < mtimes)[0]
                if indices_to_update.size > 0:
                    # Update the rows in self.df with the rows from the db
                    self.update_df_rows_from_db(
                        db, ids[indices_to_update], update_cache=update_cache
                    )
            return indices_to_remove, indices_to_update, indices_to_add
        else:
            return np.array([]), np.array([]), np.array([])

    def update_df_rows_from_db(
        self, db, indices: np.ndarray, update_cache=True
    ) -> None:
        for idx in indices:
            # If some keys from self.user_keys have been removed we
            # have to remove them from self.df before adding new
            # information, since the new information only contains the
            # user_keys that are currently present in the db
            self.df.loc[(self.df.id == idx).to_numpy(), self.user_keys] = None

            # Then get new information
            df, user_keys = db_to_df(db, sel=f"id={idx}")

            # Check that the dtypes of common columns are compatible
            for col in set(user_keys) & set(self.user_keys):
                if not is_dtypes_compatible(self.df[col].dtype, df[col].dtype):
                    # Not compatible, convert the column to object
                    self.df[col] = self.df[col].astype(object)

            self.df.loc[(self.df.id == idx).to_numpy(), ALL_COLUMNS + user_keys] = (
                df.loc[0].to_numpy()
            )

        self.clean_user_keys()

        if update_cache:
            self.clear_all_caches()

    def delete_rows_from_df_and_db(self, row_ids: Iterable[int]) -> None:
        """Delete the row id(s) from the database and self.df"""
        self.delete_rows_from_df([self.index_from_row_id(row_id) for row_id in row_ids])

        with connect(self.db_path) as db:
            db.delete(row_ids)

    def add_new_user_key(self, key: str, show=False) -> None:
        """Add a new key to the user keys. This could be when a new
        key value pair has been added.

        If show is True, the new key will also be added to chosen_columns.

        """
        if key not in self.user_keys:
            self.user_keys.append(key)
        if show:
            self.add_to_chosen_columns(key)

    def clean_user_keys(self) -> None:
        """Go through df, if a column is None in all rows,
        then remove it from user_keys and df."""
        # Cannot remove directly from df since we have some columns we
        # don't want to remove, e.g. calculator, energy, ... that
        # could be empty
        remaining_user_keys = set(
            self.df[self.user_keys].dropna(axis=1, how="all").columns
        )
        columns_to_drop = list(set(self.user_keys) - remaining_user_keys)

        # Drop from df
        self.df.drop(labels=columns_to_drop, axis=1, inplace=True)

        # Drop from chosen_columns
        for drop_column in set(columns_to_drop) & set(self.chosen_columns):
            self.remove_from_chosen_columns(drop_column)

        # Drop from user_keys
        self.user_keys = list(remaining_user_keys)

    def delete_rows_from_df(
        self, indices: Iterable[int], update_cache: bool = True
    ) -> None:
        """Delete the row id(s) from self.df.

        Remove keys from user_keys if no remaining rows have the key.

        Also take care of the cache."""
        self.df.drop(indices, inplace=True)
        self.df.reset_index(inplace=True)
        self.clean_user_keys()

        if update_cache:
            # _string_df_cache can be rebuilt quickly from
            # _string_column_cache and _filter_mask_cache. In the latter
            # two we can just remove the indices corresponding to the rows
            # that are deleted.
            self._string_df_cache.clear()
            for key in self._filter_mask_cache.keys():
                self._filter_mask_cache[key] = np.delete(
                    self._filter_mask_cache[key], indices
                )
            for key in self._string_column_cache.keys():
                self._string_column_cache[key] = self._string_column_cache[key].drop(
                    indices
                )


def ids_and_mtimes(db) -> tuple[np.ndarray, np.ndarray]:
    ids, mtimes = zip(
        *[(row.id, row.mtime) for row in db.select(columns=["id", "mtime"])]
    )
    return np.asarray(ids), np.asarray(mtimes)


def apply_filter_and_sort_on_df(
    df: pd.DataFrame, filter_mask: np.ndarray, sort: np.ndarray
) -> pd.DataFrame:
    """Apply filter mask and sorting indices on a DataFrame. Return the result.

    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame to be filtered and sorted.
    filter_mask : np.ndarray
        A boolean array of the same length as the DataFrame's index, indicating which rows to keep.
    sort : np.ndarray
        An array of integers of the same length as the DataFrame's index, indicating the order of the rows.

    Returns
    -------
    pd.DataFrame
        The filtered and sorted DataFrame.

    Examples
    --------
    >>> df = pd.DataFrame({'name': ['Alice', 'Bob', 'Charlie', 'David'], 'age': [25, 30, 35, 40]})
    >>> filter_mask = np.array([True, False, True, False])
    >>> sort = np.array([2, 0, 3, 1])
    >>> apply_filter_and_sort_on_df(df, filter_mask, sort)
          name  age
    2  Charlie   35
    0    Alice   25
    """
    return df.iloc[sort].iloc[filter_mask[sort]]


def instantiate_data(db_path: str, sel: str = "", limit: int | None = None) -> Data:
    db = connect(db_path)
    df, user_keys = db_to_df(db, sel, limit)
    return Data(df=df, db_path=Path(db_path), user_keys=user_keys)


def db_to_df(db, sel="", limit: int | None = None) -> tuple[pd.DataFrame, List[str]]:
    """Convert a db into a pandas.DataFrame.

    The columns are built using defaultdicts, and put into a
    dataframe in the end.

    """
    cols = defaultdict(list)
    user_keys = set()
    keys = ALL_COLUMNS
    i = 0
    for row in db.select(selection=sel, limit=limit):
        # default keys are always present
        for k in keys:
            cols[k].append(get_value(row, k))
        user_keys |= set(row.key_value_pairs.keys())
        for k in user_keys:
            cols[k].extend([None] * (i - len(cols[k])) + [get_value(row, k)])
        i += 1
    # Turn the lists into pd.Series so they will get the correct data
    # type from the beginning. If we use astype on the individual
    # columns we could lose information, if e.g. a column contains
    # both integers and floats.
    for k in cols.keys():
        dtype = COLUMN_DTYPES.get(k, recommend_dtype(cols[k]))
        cols[k] = pd.Series(cols[k], dtype=dtype)
    df = pd.DataFrame(cols)
    return df, list(user_keys)


def get_value(row, key) -> str:
    """Get the value from the row to the dataframe."""
    if key == "age":
        value = row.ctime
    elif key == "modified":
        value = row.mtime
    elif key == "pbc":
        # Convert pbc to a string of F and T because we don't want
        # to save arrays in the df
        value = "".join("FT"[int(p)] for p in row.pbc)
    else:
        value = row.get(key, None)
    return value


def get_data(db_path, row_id):
    db = connect(db_path)
    return db.get(id=row_id).data


def recommend_dtype(iterable):
    """Recommend a dtype for a column based on the types in the iterable.

    All recommended dtypes should be nullable, i.e. we can add None
    and keep the same dtype.

    This should only be used for the additional columns in
    user_keys. Standard columns should have a fixed dtype defined in
    COLUMN_DTYPES.

    """

    # Initialize variables to keep track of types
    has_float = False
    has_int = False
    has_str = False
    has_bool = False

    # Inspect the types in the iterable
    for item in iterable:
        if isinstance(item, bool):
            has_bool = True
        elif isinstance(item, float):  # and not item.is_integer():
            has_float = True
        elif isinstance(item, int):
            has_int = True
        elif isinstance(item, str):
            has_str = True

    # Determine the recommended dtype
    if has_str and not (has_int or has_float or has_bool):
        return pd.StringDtype()  # Only strings, return StringDtype
    elif has_bool and not (has_int or has_float or has_str):
        return pd.BooleanDtype()  # Only booleans, return BooleanDtype
    elif has_float and not has_int:
        return "float"  # Only floats, return 'float'
    elif has_float or (has_int and has_float):
        return "object"  # Mixed integers and floats, return 'object'
    elif has_int and not (has_str or has_bool):
        return pd.Int64Dtype()  # Integers, return nullable integer dtype
    else:
        return "object"  # Fallback to 'object' for other cases


def is_dtype_compatible_with_value(dtype, value):
    if dtype == "object":
        return True
    if dtype == "float":
        return isinstance(value, float) or pd.isna(value)
    if dtype == pd.BooleanDtype():
        return isinstance(value, bool) or pd.isna(value)
    if dtype == pd.StringDtype():
        return isinstance(value, str) or pd.isna(value)
    if dtype == "int":
        return isinstance(value, int) and not isinstance(value, bool)
    if dtype == pd.Int64Dtype():
        return (isinstance(value, int) and not isinstance(value, bool)) or pd.isna(
            value
        )
    return False


def is_dtypes_compatible(dtype1, dtype2):
    """Combine all possible types, return True if they are compatible,
    i.e. no upcasting with resulting data loss will take place.

    For example: int -> object is ok, but int -> float is not ok.
    """
    types = (dtype1, dtype2)
    if dtype1 == dtype2:
        return True
    if dtype1 == "object" or dtype2 == "object":
        return True
    incompatible_types = ["float", pd.BooleanDtype(), pd.StringDtype(), pd.Int64Dtype()]
    # Test all combinations of size 2 of incompatible types
    for a, b in combinations(incompatible_types, 2):
        if a in types and b in types:
            return False
    raise ValueError(f"Unknown dtype: {dtype1} or {dtype2}")



import re
from typing import Tuple

from textual.binding import Binding
from textual.containers import Container


class ColumnAdd(Container):
    pass


def valid_regex(input_string: str) -> bool:
    """Check if the input string is a valid regex pattern"""
    try:
        re.compile(input_string)
        is_valid = True
    except re.error:
        is_valid = False
    return is_valid


def check_escape_character(input_string: str) -> bool:
    """Check if the final character is an escape character."""
    char = input_string[-1]
    if char == "\\":
        return True
    return False


def ready_for_regex_search(input_string: str) -> bool:
    """Check if the input string is ready for a regex search."""
    # If the final character is an escape character, return False
    if check_escape_character(input_string):
        return False
    # If the input string is not a valid regex, return False
    if not valid_regex(input_string):
        return False
    return True


class Search(Container):
    _table = None
    _data = None
    _coordinates = []

    BINDINGS = [
        Binding("ctrl+s", "next", "Next match"),
        Binding("ctrl+r", "previous", "Previous match"),
    ]

    def set_current_cursor_coordinate(self) -> None:
        """Set the current cursor coordinate."""
        self._current_cursor_coordinate = self._table.cursor_coordinate

    def on_input_changed(self, input):
        if input.value != "":
            coordinates = self._data.search_for_string(
                input.value, regex=ready_for_regex_search(input.value)
            )

            self._coordinates = list(coordinates)

            if len(self._coordinates) > 0:
                self.move_cursor(self._coordinates[0])

    def move_cursor(self, coordinate: Tuple[int, int]) -> None:
        self._table.move_cursor(row=coordinate[0], column=coordinate[1])

    def on_input_submitted(self, input) -> None:
        """Hide the search box after pressing enter. But first clear
        the input field."""
        self.app.show_search_box = False
        self._table.focus()

    def action_next(self) -> None:
        """Move the cursor to the next match."""
        self.move_to_match()

    def action_previous(self) -> None:
        """Move the cursor to the previous match."""
        self.move_to_match(previous=True)

    def move_to_match(self, previous=False) -> None:
        """Move the cursor to the next or previous match."""
        direction = 1
        if previous:
            direction = -1

        if len(self._coordinates) == 0:
            return

        try:
            current_index = self._coordinates.index(self._table.cursor_coordinate)
        except ValueError:
            current_index = -1

        try:
            next_coordinate = self._coordinates[current_index + direction]
        except IndexError:
            next_coordinate = self._coordinates[0]

        self.move_cursor(next_coordinate)



from textual.validation import Function

from texase.formatting import convert_str_to_other_type


def contains_equals_sign(value: str) -> bool:
    return "=" in value


def not_only_whitespace(value: str) -> bool:
    for input in value.split("="):
        if not input.strip():
            return False
    return True


def no_comma(value: str) -> bool:
    return "," not in value


def can_be_interpreted_only_as_str_number_or_bool(value: str) -> bool:
    """Return True if the value can be interpreted as something else
    than a string, floating point number, integer or boolean."""
    new_value = convert_str_to_other_type(value.split("=")[-1].strip())
    return isinstance(new_value, (str, float, int, bool))


kvp_validators_edit = [
    Function(not_only_whitespace, "The key or value can't be only whitespace!"),
    Function(
        no_comma,
        "The key or value can't contain a comma! Only one key-value pair can be added at a time.",
    ),
    Function(
        can_be_interpreted_only_as_str_number_or_bool,
        "The value can be interpreted as something other than a string, number or boolean.",
    ),
]

kvp_validators_add = kvp_validators_edit + [
    Function(contains_equals_sign, "Must contain '='"),
]



from __future__ import annotations

from textual import on
from textual.app import ComposeResult
from textual.validation import ValidationResult, Validator
from textual.widgets import Input, Label

from texase.edit import EditBox
from texase.filter import FilterSuggester


class AddColumnBox(EditBox):
    input_widget_id = "add-column-input"
    label_id = "add-column-label"

    def compose(self) -> ComposeResult:
        yield Label("Add column:", id=self.label_id, classes="bottom-labels")
        yield Input(
            id=self.input_widget_id,
            classes="bottom-inputs",
            suggester=ColumnSuggester(self.app),
            validate_on=["submitted"],
            validators=[
                ValidColumn(app=self.app),
            ],
        )

    @on(Input.Submitted)
    def show_invalid_reasons(self, event: Input.Submitted) -> None:
        # Updating the UI to show the reasons why validation failed
        if not event.validation_result.is_valid:
            self.query_one(f"#{self.label_id}").add_class("-invalid")
            self.notify(
                "\n".join(event.validation_result.failure_descriptions),
                title="Invalid input",
                severity="error",
            )
        else:
            self.query_one(f"#{self.label_id}").remove_class("-invalid")


class ValidColumn(Validator):
    def __init__(self, *args, **kwargs) -> None:
        self._app = kwargs.pop("app", None)
        super().__init__(*args, **kwargs)

    def validate(self, value: str) -> ValidationResult:
        """Check if the value is an unused column."""
        if self._app.data.can_column_be_added(value):
            return self.success()
        else:
            return self.failure(f"{value} is not a valid column!")


class ColumnSuggester(FilterSuggester):
    async def get_suggestion(self, value: str) -> str | None:
        """Gets a completion from the unused columns in app.data.

        Args:
            value: The current value.

        Returns:
            A valid completion suggestion or `None`.
        """
        possible_suggestions = self._app.data.unused_columns()
        for idx, suggestion in enumerate(possible_suggestions):
            if suggestion.startswith(value):
                return possible_suggestions[idx]
        return None



import webbrowser

from textual import on
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Container, Horizontal, VerticalScroll
from textual.screen import ModalScreen
from textual.widgets import Footer, Markdown

HELP_MARKDOWN_TOP = """\
# [Texase](https://github.com/steenlysgaard/texase?tab=readme-ov-file#texase)

## Key bindings

"""

HELP_MARKDOWN_LEFT = """\
### Navigation

- `Arrow keys`: Move cursor
- `Home/End`: Go to top/bottom

### Manipulation

- `Space`: Mark/unmark row
- `u`: Unmark row
- `U`: Unmark all rows
- `#`: Delete marked rows (prompts y/n)
- `x`: Export marked rows to file (prompts for filename)
- `i`: Import rows from file (prompts for filename)

### Editing

- `e`: Edit field (if editable)
- `d`: Delete currently selected key-value-pair
- `k`: Add key-value-pair (prompts for key and value)

### More information

- `v`: View marked rows (if no rows are marked then view current row)
- `Enter`: Open details

"""

HELP_MARKDOWN_RIGHT = """\
### Filtering and searching

- `/`: Filter rows (prompts for filter string)
- `Ctrl-s`: Search rows (prompts for search string)

### Table appearance

- `s`: Sort by current column
- `+`: Add column
- `-`: Remove current selected column

### Misc

- `g`: Update table (from database)
- `Ctrl-g`: Hide all boxes, show only table
- `?`: Toggle this help
- `q`: Quit

"""

HELP_MARKDOWN_BOTTOM = """\
## Links

- [ASE DB](https://wiki.fysik.dtu.dk/ase/ase/db/db.html#module-ase.db)
- [ASE units](https://wiki.fysik.dtu.dk/ase/ase/units.html#units)
"""


class HelpScreen(ModalScreen):
    BINDINGS = [
        ("?", "pop_screen", "Toggle this help"),
        Binding("q", "pop_screen", "Hide help", show=False),
        Binding("ctrl+g", "pop_screen", "Hide help", show=False),
    ]

    DEFAULT_CSS = """
        HelpScreen {
            align: center middle;
        }

        #help-screen-container {
            width: auto;
            max-width: 80%;
            height: auto;
            max-height: 90%;
            padding: 2 4;
            align: center middle;
            background: $panel;
        }

        Container > VerticalScroll > Horizontal > Markdown {
            width: 50%;
        }

        Container > VerticalScroll > Horizontal {
            height: auto;
        }
"""

    def compose(self) -> ComposeResult:
        with Container(id="help-screen-container"):
            with VerticalScroll():
                yield Markdown(HELP_MARKDOWN_TOP, id="help-markdown-top")
                with Horizontal():
                    yield Markdown(HELP_MARKDOWN_LEFT, id="help-markdown-left")
                    yield Markdown(HELP_MARKDOWN_RIGHT, id="help-markdown-right")
                yield Markdown(HELP_MARKDOWN_BOTTOM, id="help-markdown-bottom")
        yield Footer()

    @on(Markdown.LinkClicked)
    def open_url(self, event: Markdown.LinkClicked) -> None:
        webbrowser.open(event.href)
        event.stop()



from __future__ import annotations

from typing import Any

from rich.text import Text
from textual import on, work
from textual.app import ComposeResult
from textual.containers import Container, Horizontal
from textual.message import Message
from textual.reactive import reactive
from textual.widgets import Input, Label, ListItem, ListView

from texase.formatting import (
    convert_str_to_other_type,
    convert_value_to_int_float_or_bool,
    get_age_string,
)
from texase.input_screens import DataEditScreen, DataScreen, KVPEditScreen, KVPScreen
from texase.validators import kvp_validators_edit


class Details(Container):
    BINDINGS = [
        # Hiding boxes shortcut is already defined in app, just
        # redefining the binding here to make it show in the edit
        # box
        ("ctrl+g", "hide_edit", "Undo/Hide"),
        ("ctrl+s", "save", "Save changes"),
        ("ctrl+d", "delete", "Delete"),
        ("k", "add_kvp", "Add key value pair"),
        ("d", "add_data", "Add data"),
    ]

    modified_keys = set()
    deleted_keys = set()
    # Watch if anything is modified. If so, show a label that says unsaved changes.
    anything_modified = reactive(False)  # TODO: Can this be changed to var??

    def compose(self) -> ComposeResult:
        yield Title(
            "Key value pairs ([@click=\"app.open_link('https://wiki.fysik.dtu.dk/ase/ase/units.html')\"]Units[/])"
        )
        yield Container(KVPStatic(), KVPList(id="dynamic_kvp_list"), id="kvp")
        yield Title("Data")
        yield DataList(id="datalist")
        yield Label("Unsaved changes!", id="unsaved_changes")

    def watch_anything_modified(self, modified):
        if modified:
            self.query_one("#unsaved_changes", Label).remove_class("-hidden")
        else:
            self.query_one("#unsaved_changes", Label).add_class("-hidden")

    def set_focus(self) -> None:
        # Remember where old focus was and start from this. If this
        # row hasn't been focused before then focus on the KVPList.
        self.query_one("#dynamic_kvp_list", KVPList).focus()

    def update_kvplist(self, static_kvps: dict, dynamic_kvps: dict) -> None:
        """Update the kvp widgets."""
        # Static (non-editable) key value pairs
        static_str = ""
        for key, value in static_kvps.items():
            if key == "age":
                value = get_age_string(value)
            static_str += f"[bold]{key}: [/bold]{value}\n"
        self.query_one(KVPStatic).update(Text.from_markup(static_str))

        # Dynamic (editable) key value pairs
        kvp_widget = self.query_one("#dynamic_kvp_list", KVPList)
        kvp_widget.clear()
        for key, value in dynamic_kvps.items():
            kvp_widget.append(ListItem(EditableItem(key, value)))

    def update_data(self, dynamic_data: dict) -> None:
        data_widget = self.query_one(DataList)
        data_widget.clear()
        for key, value in dynamic_data.items():
            data_widget.append(ListItem(DataItem(key, value)))

    def on_hide(self) -> None:
        self.clear_modified_and_deleted_keys()

    def clear_modified_and_deleted_keys(self) -> None:
        self.modified_keys.clear()
        self.deleted_keys.clear()
        self.anything_modified = False

    def on_details_list_item_selected(self, sender: DetailsList.ItemSelected):
        self.add_key_to_modified(sender.item.key)

    def add_key_to_modified(self, key: str) -> None:
        self.modified_keys.add(key)
        self.anything_modified = True

    def action_save(self) -> None:
        """Save the changes to the table, dataframe and database."""

        # Get the key value pairs from the KVPList
        key_value_pairs = {}
        for h in self.query_one("#dynamic_kvp_list", KVPList).children:
            item = h.get_child_by_type(EditableItem)
            key = item.key
            if key in self.modified_keys:
                key_value_pairs[key] = item.value

        # Do the same for updated data
        updated_data = {}
        for h in self.query_one("#datalist", DataList).children:
            item = h.get_child_by_type(DataItem)
            key = item.key
            if key in self.modified_keys:
                updated_data[key] = item.value

        self.app.save_details(key_value_pairs, updated_data, self.deleted_keys)
        self.clear_modified_and_deleted_keys()

    def action_delete(self) -> None:
        """Delete the current key value pair. If the selection is on
        the DataList then notify that ASE can't delete data items in a
        row."""
        focused_list: DetailsList = self.app.focused
        if isinstance(focused_list, DataList):
            self.notify("ASE can't delete data items in a row.", severity="information")
            return
        selected_key = focused_list.selected_key()
        if selected_key is not None:
            self.deleted_keys.add(selected_key)
            self.anything_modified = True
        focused_list.delete_selected()

    @work
    async def action_add_data(self) -> None:
        """Add a new data item."""
        new_data = await self.app.push_screen_wait(DataScreen("Add data"))
        if new_data is not None:
            key, value = new_data
            self.query_one(DataList).append(ListItem(DataItem(key, value)))
            self.add_key_to_modified(key)

    @work
    async def action_add_kvp(self) -> None:
        """Add a new key value pair item."""
        new_kvp = await self.app.push_screen_wait(KVPScreen("Add key value pair"))
        if new_kvp is not None:
            key, value = new_kvp
            self.query_one(KVPList).append(ListItem(EditableItem(key, value)))
            self.add_key_to_modified(key)

    @work
    async def on_details_list_open_input_screen_with_item(
        self, sender: DetailsList.OpenInputScreenWithItem
    ):
        item = sender.item
        EditScreen = KVPEditScreen if isinstance(item, EditableItem) else DataEditScreen
        new_value = await self.app.push_screen_wait(
            EditScreen(
                "Edit key value pair", key=item.key, prefilled_input=str(item.value)
            )
        )
        if new_value is not None:
            item.value = new_value
            self.add_key_to_modified(item.key)


class Item(Horizontal):
    validators = []

    _value: Any = None

    def __init__(self, key, value):
        super().__init__()
        self.key = key
        self.value = value

    @property
    def value(self: Any) -> None:
        return self._value

    @value.setter
    def value(self, value: Any) -> None:
        # After changing from None only set the value
        if self._value is not None:
            self.app.notify_if_kvp_type_changed(self.key, value, self._value)
            self.query_one(Input).value = str(value)
        self._value = value

    def compose(self) -> ComposeResult:
        yield Label(f"[bold]{self.key} = [/bold]")
        yield Input(
            value=f"{self.value}",
            classes="editable-input",
            validators=self.validators,
            validate_on=["submitted"],
        )

    def focus(self) -> None:
        self.query_one(Input).focus()


class EditableItem(Item):
    validators = kvp_validators_edit

    def on_input_submitted(self, submitted: Input.Submitted):
        """When the user presses enter in the input field, update the value.

        Then the KVPList takes back focus.
        """
        # TODO: Exactly the same code as in app.py. Refactor?
        if (
            submitted.validation_result is not None
            and not submitted.validation_result.is_valid
        ):
            self.app.notify_error(
                "\n".join(submitted.validation_result.failure_descriptions),
                error_title="Invalid input",
            )
            # If not valid input stop bubbling further
            submitted.stop()
            return

        if self.key == "pbc":
            value = submitted.value.upper()
        else:
            value = convert_value_to_int_float_or_bool(submitted.value)

        if not self.app.is_kvp_valid(self.key, value):
            submitted.stop()  # stop bubbling further
            return

        self.value = value


class DataItem(Item):
    def on_input_submitted(self, submitted: Input.Submitted):
        """When the user presses enter in the input field, update the value.

        Then the DataList takes back focus.
        """
        if (
            submitted.validation_result is not None
            and not submitted.validation_result.is_valid
        ):
            self.app.notify_error(
                "\n".join(submitted.validation_result.failure_descriptions),
                error_title="Invalid input",
            )
            # If not valid input stop bubbling further
            submitted.stop()
            return

        value = convert_str_to_other_type(submitted.value)
        self.value = value


class Title(Label):
    pass


class KVPStatic(Label):
    pass


class DetailsList(ListView):
    class ItemSelected(Message):
        """Message that this item was selected."""

        def __init__(self, item: Item) -> None:
            self.item = item
            super().__init__()

    class OpenInputScreenWithItem(Message):
        """An input screen should be opened with this item as prefilled input."""

        def __init__(self, item: Item) -> None:
            self.item = item
            super().__init__()

    def on_list_view_selected(self, sender):
        """When a row is selected in the KVPList, focus on the input
        field and remember that this key value pair was potentially
        modifed. If the width of the details list is less than the
        width of the kvp, so it can't show fully, use an InputScreen
        instead

        """
        item = sender.item.children[0]
        if sender.item.container_size.width < len(item.key) + len(str(item.value)) + 3:
            self.post_message(self.OpenInputScreenWithItem(item))
            return
        item.focus()

        self.post_message(self.ItemSelected(item))

    @on(Input.Submitted)
    def take_back_focus(self, _: Input.Submitted):
        """When the user presses enter in the input field, take back
        focus. It is assumed that the input value is valid."""
        self.focus()

    def delete_selected(self) -> None:
        """Delete the current key value pair."""
        current_index = self.index
        if self.highlighted_child is not None:
            self.highlighted_child.remove()
        self.index = current_index

    def selected_key(self) -> str | None:
        raise NotImplementedError("selected_key must be implemented in subclass.")


class KVPList(DetailsList):
    def selected_key(self) -> str | None:
        """Return the key of the currently selected key value pair."""
        if self.highlighted_child is not None:
            return self.highlighted_child.get_child_by_type(EditableItem).key


class DataList(DetailsList):
    def selected_key(self) -> str | None:
        """Return the key of the currently selected key value pair."""
        if self.highlighted_child is not None:
            return self.highlighted_child.get_child_by_type(DataItem).key

    def delete_selected(self) -> None:
        raise NotImplementedError("ASE can't delete data items in a row.")



from __future__ import annotations

from rich.text import Text
from textual.app import ComposeResult
from textual.containers import ScrollableContainer
from textual.suggester import Suggester
from textual.validation import Function
from textual.widgets import Button, Checkbox, Input, Select, Static
from textual.widgets._data_table import RowKey

from texase.data import ops
from texase.table import TexaseTable


class FilterBox(ScrollableContainer):
    BINDINGS = [
        # Hiding boxes shortcut is already defined in app, just
        # redefining the binding here to make it show in the filter
        # box
        ("ctrl+g", "hide_filter", "Hide filter(s)"),
        ("ctrl+t", "toggle_only_marked", "Toggle only marked"),
        ("ctrl+n", "add_filter", "New filter"),
        ("ctrl+r", "remove_filter", "Remove filter"),
    ]

    async def focus_filterbox(self) -> None:
        """Handle putting focus on the filterbox.

        If there are no filters, add a filter and put focus on the filterkey of that filter.
        If there are filters, go to the last filter then
            if the filter is disabled put focus on the + button
            else put focus on the filterkey.
        """

        filters = self.query("#filterkey")
        if len(filters) == 0:
            await self.add_filter()
            return

        # There could be more filters, so we focus on the last one
        if filters[-1].disabled:
            # The filter is disabled we focus on the + button instead
            self.query("#add-filter")[-1].focus()
        else:
            filters[-1].focus()

    def action_toggle_only_marked(self) -> None:
        """Toggle the only marked checkbox."""
        checkbox = self.query_one("#only-marked-checkbox")
        checkbox.value = not checkbox.value

    async def action_add_filter(self) -> None:
        await self.add_filter()

    async def add_filter(self) -> Filter:
        new_filter = Filter()
        await self.mount(new_filter)
        new_filter.scroll_visible()
        self.query("#filterkey")[-1].focus()
        return new_filter

    def action_remove_filter(self) -> None:
        """Remove the currently focused filter, alternatively the last
        filter.

        """
        filters = self.query(Filter)
        for filter in filters:
            if filter.has_focus_in_any_widget():
                filter.remove()
                break
        else:
            # If no filter has focus, remove the last filter
            if len(filters) > 0:
                filters[-1].remove()
        self.hide_if_no_filters()

    def compose(self) -> ComposeResult:
        yield Filter()
        yield Checkbox("Only mark filtered", id="only-marked-checkbox")

    async def on_button_pressed(self, pressed: Button.Pressed) -> None:
        """Called when a button is pressed."""
        if pressed.button.id == "remove-filter":
            # Remove the filter
            pressed.button.ancestors[0].remove()
            self.hide_if_no_filters()
        elif pressed.button.id == "add-filter":
            await self.add_filter()

    def hide_if_no_filters(self) -> None:
        """If there are no more filters hide the filter box."""
        if len(self.query(Filter)) == 0:
            self.app.show_filter = False


class Filter(Static):
    applied = False

    def on_input_submitted(self, submitted: Input.Submitted) -> None:
        """Called when the user presses enter on the filtervalue input."""
        input = submitted.input
        # Maybe not a great way to get the table
        app = self.ancestors[-1]
        if input.id == "filterkey":
            if input.value not in app.data.chosen_columns:
                self.notify(
                    f"Column {input.value} does not exist",
                    severity="warning",
                    title="Warning",
                )
            return

        key_input = self.query_one("#filterkey")
        op = self.query_one("#filteroperator")

        table = app.query_one(TexaseTable)
        filterbox = self.ancestors[0]
        if filterbox.query_one("#only-marked-checkbox").value:
            # Only mark
            filter_mask = app.data.get_mask_of_df_with_filter(
                (key_input.value, op.value, input.value)
            )
            for i in app.data.id_array_with_filter_and_sort(filter_mask):
                table.mark_row(RowKey(str(i)))
        else:
            # Filter
            table.add_filter(key_input.value, op.value, input.value)

            # Upon succesfully applying the filter, set the applied
            # variable to True
            self.applied = True

            # Disable the input widgets so they can't be changed, then it
            # is easier to remove them again
            input.disabled = True
            op.disabled = True
            key_input.disabled = True

        # Hide yourself
        app.show_filter = False

        # Focus on the table
        table.focus()

    def remove(self) -> None:
        # If this filter is applied to the data shown in the
        # table, it should be removed there as well

        if self.applied:
            key = self.query_one("#filterkey").value
            op = self.query_one("#filteroperator").value
            val = self.query_one("#filtervalue").value
            self.app.remove_filter_from_table((key, op, val))
        # After this the parent container takes care of removing this filter

        super().remove()

    def compose(self) -> ComposeResult:
        yield Button(
            Text("\uff0b", style="green"),
            id="add-filter",
            classes="filter-buttons",
            variant="primary",
        )
        yield Input(
            placeholder="Filter on column..",
            suggester=ColumnSuggester(self.ancestors[-1]),
            validate_on=["blur"],
            validators=[
                Function(
                    lambda x: x in self.app.data.chosen_columns, "Column does not exist"
                )
            ],
            id="filterkey",
        )
        yield Select(
            [(op, op) for op in ops.keys()],
            value="==",
            allow_blank=False,
            id="filteroperator",
        )
        yield Input(placeholder="Value", id="filtervalue")
        yield Button(
            Text("\u2212", style="red"),
            id="remove-filter",
            classes="filter-buttons",
            variant="warning",
        )

    def has_focus_in_any_widget(self) -> bool:
        """Return True if any widget in this container has focus."""
        for selector in [
            "#add-filter",
            "#filterkey",
            "#filteroperator",
            "#filtervalue",
            "#remove-filter",
        ]:
            if self.query_one(selector).has_focus:
                return True
        return False


class FilterSuggester(Suggester):
    """Give completion suggestions based either on columns or values
    of a single column."""

    def __init__(self, app, *, case_sensitive: bool = True) -> None:
        """Creates a suggester based off of a given iterable of possibilities.

        Args:
            suggestions: Valid suggestions sorted by decreasing priority.
            case_sensitive: Whether suggestions are computed in a case sensitive manner
                or not. The values provided in the argument `suggestions` represent the
                canonical representation of the completions and they will be suggested
                with that same casing.
        """
        super().__init__(case_sensitive=case_sensitive)
        self._app = app


class ColumnSuggester(FilterSuggester):
    async def get_suggestion(self, value: str) -> str | None:
        """Gets a completion from the given possibilities.

        Args:
            value: The current value.

        Returns:
            A valid completion suggestion or `None`.
        """
        possible_suggestions = self._app.data.chosen_columns
        for idx, suggestion in enumerate(possible_suggestions):
            if suggestion.startswith(value):
                return possible_suggestions[idx]
        return None
